import logging
import optparse
import os
import sys
import yaml
from zonetool import zone


def main():
    parser = optparse.OptionParser(
        usage='%prog [<OPTIONS>] <PATH>...')
    parser.add_option(
        '-c', '--config', dest='config', metavar='FILE',
        help='Load global configuration from FILE (in YAML format)')
    parser.add_option(
        '--named-conf', dest='named_conf', metavar='FILE',
        help='Location of the autogenerated named.conf snippet '
        '(the file will not be created if the option is unset)')
    parser.add_option(
        '-o', '--output-dir', dest='output_dir',
        default='.', metavar='DIR',
        help='Output directory (default: current dir)')
    parser.add_option(
        '--delete', dest='delete', action='store_true',
        help='Delete obsolete zone files in output_dir')
    parser.add_option(
        '-n', '--dry-run', dest='dry_run', action='store_true',
        help='Do not actually write zone files')
    parser.add_option(
        '--update-policy', dest='update_policies', action='append',
        help='Set the update-policy for all zones (specify more than once for multiple policies)')
    parser.add_option(
        '--debug', action='store_true',
        help='Show debugging output')

    dnssec_g = parser.add_option_group('DNSSEC Options')
    dnssec_g.add_option(
        '--key-dir', dest='key_dir', metavar='DIR',
        help='Location of key files')
    dnssec_g.add_option(
        '--ds-dir', dest='ds_dir', metavar='DIR',
        help='Location of DS record files')
    dnssec_g.add_option(
        '--nsec3-salt', dest='nsec3_salt', metavar='HEX',
        help='NSEC3 salt (hex-encoded)')
    dnssec_g.add_option(
        '--dnssec-refresh', dest='dnssec_refresh', action='store_true',
        help='Refresh all DNSSEC signatures')

    opts, args = parser.parse_args()
    if not args:
        parser.error('Not enough arguments')

    logging.basicConfig(level=logging.DEBUG if opts.debug else logging.INFO)

    # Create a global config dictionary using the environment and (if
    # specified) a YAML config file. The advantage of the config file
    # is that it lets you define variables which are not just plain
    # strings.
    config = dict(os.environ)
    if opts.config:
        with open(opts.config, 'r') as fd:
            config.update(yaml.safe_load(fd))

    # Parse the DNS configuration and generate zone data.
    zp = zone.ZoneParser(config)
    for arg in args:
        zp.load(zone.readyaml(zone.walk(arg)))

    pproc = None
    if opts.key_dir and opts.ds_dir and opts.nsec3_salt:
        # Enable DNSSEC support.
        pproc = zone.DNSSECSigner(
            opts.key_dir, opts.ds_dir, opts.nsec3_salt, opts.dnssec_refresh, opts.dry_run)

    # Render the zone data to 'output_dir'.
    zw = zone.ZoneWriter(opts.output_dir,
                         opts.named_conf,
                         delete=opts.delete,
                         dry_run=opts.dry_run,
                         update_policies=opts.update_policies)
    changed, removed = zw.write(zp.render(), postproc=pproc)
    sys.exit(0)


if __name__ == '__main__':
    main()
