import glob
import logging
import os
import re
import socket
import subprocess
import time
import yaml


# List of known zone attributes.
ATTRIBUTES = frozenset(['EXTENDS', 'REPLACES', 'DNSSEC'])


class Error(Exception):
    pass


def _is_ip(s, family):
    try:
        socket.inet_pton(family, s)
        return True
    except:
        return False


def _is_ip6(s):
    return _is_ip(s, socket.AF_INET6)


def _is_ip4(s):
    return _is_ip(s, socket.AF_INET)


_COMMENTS_RE = re.compile(r';[^\n]*$')
_SPACES_RE = re.compile(r'\s+')
_NOSERIAL_RE = re.compile(r'^\s*(\d+)\s*;\s*Serial(|\s.*)$')

def zonecmp(a, b):
    """Check if two zones are equal.

    Ignores SOA serial number, comments, and whitespace.
    """
    def _lines(z):
        return [y for y in [
            _COMMENTS_RE.sub('', _SPACES_RE.sub(' ', _NOSERIAL_RE.sub('', x.strip())))
            for x in z.split('\n')] if y]
    aa = '\n'.join(_lines(a))
    bb = '\n'.join(_lines(b))
    return aa == bb


def _parse_record(value):
    # Auto-detect A and AAAA record types.
    if _is_ip6(value):
        rtype, rdata = ('AAAA', value)
    elif _is_ip4(value):
        rtype, rdata = ('A', value)
    else:
        try:
            rtype, rdata = value.split(None, 1)
        except ValueError:
            raise Error('invalid record value "%s"' % value)
    return rtype.upper(), rdata


def _to_records(data):
    records = []
    for key, values in data.items():
        # Skip zone attributes when generating records.
        if key in ATTRIBUTES:
            continue
        if isinstance(values, str):
            values = [values]
        elif not isinstance(values, list):
            # Ignore values of other types (bools?)
            continue
        for value in values:
            rtype, rdata = _parse_record(value)
            records.append(('' if key == '_' else  key, rtype, rdata))
    return sorted(records)


def _render_zone(name, data, attrs):
    records = _to_records(data)
    nameservers = [x[2] for x in records if x[0] == '' and x[1] == 'NS']
    if not nameservers:
        nameservers = ['localhost.']

    preamble = '''; Autogenerated, DO NOT EDIT!
$TTL %(ttl)d
@\t\tIN\tSOA\t%(first_nameserver)s\thostmaster.%(origin)s. (
\t\t\t\t%(serial)d ; Serial
\t\t\t\t%(refresh_ttl)d
\t\t\t\t%(retry_ttl)d
\t\t\t\t%(expire_ttl)d
\t\t\t\t%(neg_cache_ttl)d )
''' % {'origin': name,
       'serial': attrs.get('SERIAL', int(time.time())),
       'ttl': attrs.get('TTL', 3600),
       'refresh_ttl': attrs.get('REFRESH', 43200),
       'retry_ttl': attrs.get('RETRY', 3600),
       'expire_ttl': attrs.get('EXPIRE', 2419200),
       'neg_cache_ttl': attrs.get('NEG_CACHE', 3600),
       'first_nameserver': nameservers[0]}

    out = [preamble]
    for key, rtype, rdata in records:
        out.append('%s\t\tIN\t%s\t%s\n' % (key, rtype, rdata))
    return ''.join(out)


def _mklist(v):
    if v is None:
        return []
    return v if isinstance(v, list) else [v]


def _unlist(v):
    if len(v) == 1:
        return v[0]
    return v


def _uniq(l):
    return list(set(l))


def _merge_zones(a, b, replace):
    """Merge two zones (dictionaries), adding b to a.

    Adds all the attributes of b to a. When a already has a given
    attribute, what happens depends on the chosen mode: if replace is
    True, the new value replaces the old one, otherwise they are
    merged into a list.

    Returns a (which is modified in-place).
    """
    for key in b:
        if not replace:
            v = _mklist(a.get(key))
            v.extend(_mklist(b[key]))
            a[key] = _unlist(_uniq(v))
    return a


_forced_paths = ['/usr/bin', '/usr/sbin']

def _fix_path():
    environ = os.environ.copy()
    path = environ['PATH'].split(os.path.pathsep)
    if not path:
        path = []
    for p in _forced_paths:
        if p not in path:
            path.append(p)
    environ['PATH'] = os.path.pathsep.join(path)
    return environ


def _exec(*cmd):
    cmd_str = ' '.join(cmd)
    logging.debug('running %s', cmd_str)
    p = subprocess.Popen(
        cmd,
        env=_fix_path(),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = p.communicate()
    if p.returncode != 0:
        logging.error('command %s failed with status %d\n%s',
                      cmd_str, p.returncode, stderr)
        raise Exception('execution of "%s" failed with status=%d' % (
            cmd_str, p.returncode))
    return stdout


class ZoneParser(object):

    def __init__(self, config=None):
        self.config = config or {}
        self.zones = {'@default': {}}
        self.resolved = {}

    def load(self, yamls):
        for objs in yamls:
            for y in objs:
                logging.debug('loading %s', y)
                self._merge(y)

    def _merge(self, data):
        for zone_name, zone_data in data.items():
            _merge_zones(
                self.zones.setdefault(zone_name, {}),
                zone_data,
                False)

    def render(self):
        for zone_name, zone_data in self.zones.items():
            if zone_name.startswith('@'):
                continue
            resolved = self._resolve_references(zone_name, zone_data)
            output = _render_zone(zone_name, resolved, resolved)
            #zone_attrs = dict((k, v) for k, v in resolved.items() if k in ATTRIBUTES)
            yield (zone_name, resolved, output)

    def _resolve_references(self, zone_name, zone_data):
        if zone_name in self.resolved:
            return self.resolved[zone_name]

        out = dict(zone_data)

        # All non-template zones implicitly inherit from @default.
        extends = ['@default'] if not zone_name.startswith('@') else []
        extends.extend(_mklist(zone_data.get('EXTENDS', [])))
        extends = _uniq(extends)
        replaces = _mklist(zone_data.get('REPLACES', []))

        # Merge extends/replaces templates into 'out'.
        for bases, replace in ((extends, False), (replaces, True)):
            for base in bases:
                if base not in self.zones:
                    raise Error('undefined template "%s"' % base)
                _merge_zones(out, self._resolve_references(
                    base, self.zones[base]), replace)

        out = self._expand_vars(out)
        out.pop('EXTENDS', None)
        out.pop('REPLACES', None)
        self.resolved[zone_name] = out
        return out

    def _expand_vars(self, zone_data):
        # This will break horribly if you use more than one
        # variable substitution per record (but why would you?)
        var_re = re.compile(r'\$([A-Za-z0-9_]+)')
        def _get_var(s):
            m = var_re.search(s)
            if m:
                try:
                    return self.config[m.group(1)], True
                except KeyError:
                    raise Error('undefined config variable "%s"' % m.group(1))
            return None, False

        def _expand_value(value):
            var_value, has_var = _get_var(value)
            if not has_var:
                return value
            elif not var_value:
                return []
            elif isinstance(var_value, list):
                return [var_re.sub(x, value) for x in var_value]
            else:
                return var_re.sub(var_value, value)

        out_zone_data = {}
        for key, value in zone_data.items():
            if isinstance(value, str):
                value = _expand_value(value)
            elif isinstance(value, list):
                tmp = []
                for v in value:
                    tmpv = _expand_value(v)
                    if tmpv:
                        if isinstance(tmpv, list):
                            tmp.extend(tmpv)
                        else:
                            tmp.append(tmpv)
                value = tmp
            # Something that expands to an empty value must be
            # removed from the zone.
            if value:
                out_zone_data[key] = value
        return out_zone_data


def walk(root_dir):
    for dirpath, dirnames, filenames in os.walk(root_dir):
        for f in filenames:
            if f.endswith('.yml') and not f.startswith('.'):
                yield os.path.join(dirpath, f)


def readyaml(filenames):
    for f in filenames:
        with open(f) as fd:
            yield yaml.safe_load_all(fd)


class ZoneWriter(object):

    def __init__(self, output_dir, named_conf, delete=False, dry_run=False,
                 update_policies=None):
        self.output_dir = output_dir
        self.named_conf = named_conf
        self.dry_run = dry_run
        self.delete_old_files = delete
        self.update_policies = update_policies

    def write_if_changed(self, data, dst):
        if os.path.exists(dst):
            with open(dst, 'r') as fd:
                if zonecmp(fd.read(), data):
                    return False
        if not self.dry_run:
            with open(dst, 'w') as fd:
                fd.write(data)
        return True

    def config_snippet(self, zone_name, zone_file):
        s = '''
zone "%s" {
  type master;
  file "%s";
  allow-query { any; };
''' % (zone_name, zone_file)
        if self.update_policies:
            for p in self.update_policies:
                s += '  update-policy { %s; };\n' % (p,)
        s += '};\n'
        return s

    def write(self, zones, postproc=None):
        # Compare the current contents of 'output_dir' with the
        # generated output to detect zones that were removed.
        existing = set(os.listdir(self.output_dir))
        changed = set()
        config = []
        for zone_name, zone_attrs, zone_data in zones:
            outfile = zone_name
            existing.discard(outfile)
            zone_changed = False
            dst = os.path.join(self.output_dir, outfile)
            if self.write_if_changed(zone_data, dst):
                logging.info('updated zone %s', zone_name)
                changed.add(zone_name)
                zone_changed = True
            # Zone file post-processing.
            if postproc:
                outfile = postproc(
                    self.output_dir, outfile, zone_name, zone_attrs, zone_changed)
                existing.discard(outfile)
            # Generate config snippet.
            config.append(self.config_snippet(
                zone_name, os.path.join(self.output_dir, outfile)))

        # Wipe what shouldn't be there.
        if self.delete_old_files:
            for filename in existing:
                logging.debug('removing %s', filename)
                if not self.dry_run:
                    os.remove(os.path.join(self.output_dir, filename))

        # Write the autogenerated named.conf.
        if not self.dry_run and self.named_conf:
            logging.debug('updated named.conf')
            with open(self.named_conf, 'w') as fd:
                for c in config:
                    fd.write(c)

        return changed, existing


class DNSSECSigner(object):

    def __init__(self, key_dir, ds_dir, nsec3_salt, refresh, dry_run=False):
        self.key_dir = key_dir
        self.ds_dir = ds_dir
        self.nsec3_salt = nsec3_salt
        self.refresh = refresh
        self.dry_run = dry_run

    def has_keys(self, zone_name):
        # One day we'll figure out what the magic numbers are...
        return glob.glob(os.path.join(self.key_dir, 'K%s.+007+*' % zone_name))

    def create_keys(self, zone_name):
        old_umask = os.umask(0o077)
        try:
            ksk_file = _exec(
                'dnssec-keygen',
                '-a', '7', '-3', '-q', '-f', 'KSK',
                '-b', '2048', '-K', self.key_dir,
                zone_name)
            zsk_file = _exec(
                'dnssec-keygen',
                '-a', '7', '-3', '-q',
                '-b', '2048', '-K', self.key_dir,
                zone_name)
            logging.info('created DNSSEC keys for "%s" (%s, %s)',
                         zone_name, ksk_file, zsk_file)
        finally:
            os.umask(old_umask)

    def sign_zone_file(self, zone_name, infile, outfile):
        """Sign a zone file with DNSSEC records.

        :param zone_name: Zone name (origin).
        :param infile: Source zone file path.
        :param outfile: Destination zone file path.
        """
        # Use only one CPU/thread on purpose otherwise the record order is
        # not preserved and the results are difficult to compare.
        logging.debug('signing zone %s', zone_name)
        _exec(
            'dnssec-signzone',
            '-a', '-g', '-S',
            '-n', '1',
            '-3', self.nsec3_salt,
            '-K', self.key_dir,
            '-d', self.ds_dir,
            '-o', zone_name,
            '-f', outfile,
            infile)

    def __call__(self, output_dir, filename, zone_name, zone_attrs, zone_changed):
        if not zone_attrs.get('DNSSEC'):
            return filename
        outfile = filename + '.signed'
        if (zone_changed or self.refresh) and not self.dry_run:
            if not self.has_keys(zone_name):
                self.create_keys(zone_name)
            self.sign_zone_file(zone_name,
                                os.path.join(output_dir, filename),
                                os.path.join(output_dir, outfile))
        return outfile
